/*
 Stephan Kottler
 kottlers@informatik.uni-tuebingen.de
 *
 */

#include "BasicSolver.hpp"
#include <algorithm>
#include "Settings.hpp"


BasicSolver::BasicSolver(Settings& sets, Statistics& stats,OFStream* out):
	ConstraintDB				(sets, stats,out),
	act_hp						(sets.var_max_act,sets.var_min_inc, sets.var_decay,
									 sets.cat_max_act,sets.cat_min_inc, sets.cat_decay),

	state_ok					( true  ),
	queue_begin					( 0	),
	simplify_assign_idx			( 0	),
	simplify_props_remaining	( 0	),
	//clause_stamp				( 0 	),
	curr_max_conflicts			( sets.conflict_max_init),
	curr_max_learnts			( 0		),
	/*curr_rand_var_freq			(sets.random_var_freq),
	curr_rand_cat_freq			(sets.random_cat_freq),*/
	reduced_learnts				(false),
	learnts_checked_to			( 0 )
{ }

BasicSolver::~BasicSolver(){}

TrBool BasicSolver::isSAT(  double timeout){

	var_flags.ensureSize(numberVariables());

	if(!state_ok)			return FALSE_TrBool;

	curr_max_learnts 	= numberAllClauses() * 2;

	TrBool status = UNKNOWN_TrBool;

	if(console_out != NULL ){
		console_out->writeString(
				"|                           SApperloT ... YOU GOT ME SEARCHING ... ");
		console_out->writeCenteredToBG(basename(settings.tgf_file_name));
		statistics.printStatsProgressHead(console_out);
		console_out->writeString(
				"|_________________________________________________________________________________________________|\n");
	}

	status =  status.isKnown() ? status : clsComputeBinTernClosure();
	if(console_out != NULL ){console_out->writeString("| ... Computed binary closure");}

	status = status.isKnown() ? status : checkSubsumed(fix_cls);
	status = status.isKnown() ? status : checkSubsumed(learnt_cls);
	if(status.isKnown()){timeout = 0;}
	if(console_out != NULL ){console_out->writeString("| ... Finished subsumption test for clauses ");}
	int next_check_forced = 20;

	//print_out("Skip subsumption test \n");
	status = status.isUnknown() && (settings.preprocess_level != 0)?(next_check_forced=0,preProcess()):status;
	//TrBool  status			= noicySearching(10);
	if(status.isKnown()){timeout = 0;}

	if(console_out != NULL ){console_out->writeString("| ... Finished preprocessing");}

	uint bin_clses = numberBinClses() + settings.bin_cls_to_force_closure;

	while(timeout < 0 || statistics.cpuUserTime() <= timeout){


		if( next_check_forced-- <= 0 || bin_clses < numberBinClses() || reduced_learnts ){

			bin_clses = numberBinClses() + settings.bin_cls_to_force_closure;

			status= clsComputeBinTernClosure();
			if(status.isKnown()) break;
			status = checkSubsumed(fix_cls);
			if(status.isKnown()) break;
			status = checkSubsumed(learnt_cls);
			if(status.isKnown()) break;



			reduced_learnts = false;
		}

		++statistics.nbr_restarts ;

		status		= search();

		if(console_out != NULL){ statistics.printStatsProgress(console_out, numberLiterals());}

		if(status.isKnown()) break;

		if(statistics.nbr_restarts % settings.restart_reset != 0){
			curr_max_conflicts   *= settings.conflict_max_inc;
		}else{
			curr_max_conflicts   = settings.conflict_max_init;
		}


		//curr_max_learnts	*= settings.learntsize_inc;

		backjumpKeepLevel(0,BACKJUMP_SET_PREV);

	}

	if(status.isUnknown()){
		assert(timeout >= 0 && timeout <= statistics.cpuUserTime());
		if(console_out != NULL)
			console_out->writeString("|!!! Timeout of %i seconds exceeded !!! ",(int)timeout);
			return UNKNOWN_TrBool;
	}

	if(console_out != NULL){
		console_out->writeString("|____________________________________________________________________________________________");
		console_out->writeString("| ");
	}

	assert(!status.isTrue() || debugVerifyModel());

	return status;
}


bool BasicSolver::addClause(vector<Literal>& lits, bool redundant){

	assert(decisionLevel() == 0);

	if (!state_ok || lits.empty()){								return false;}

	// Check if clause is satisfied and remove false/duplicate literals:
	std::sort(lits.begin(),lits.end());
	reserveForVariables(lits.back().toVar().idx()+1);

	Literal						last	(INVALID_LIT);
	uint32_t					i	(0);
	vector<Literal>::iterator   to	(lits.begin());
	while(i < lits.size()){
		if(litIsTrue(lits[i]) || lits[i] == last.flipped()){	return true;}

		if(litIsUnknown(lits[i]) && lits[i] != last){
			*to++ = last = lits[i];
			act_hp.addElementIfNotExists(last.toVar());
		}

		++i;
	}

	return addValidClause(lits.begin(),to, redundant);
}




bool BasicSolver::addValidClause(vector<Literal>::iterator lbeg, const vector<Literal>::iterator lend,bool redundant){

	assert(decisionLevel() == 0);
	uint32_t l_sz(lend-lbeg);

	if(l_sz < 4){
		if(l_sz == 1){
			assert(litIsUnknown(*lbeg));
			uncheckedEnqueue<Clause*>(*lbeg);
			//act_hp.bump(lbeg->toVar());
																	return state_ok = (bcp(false) == BCP_SUCCESS);}

		Literal lit0 = *lbeg; Literal lit1 = *(lbeg+1);
		if(l_sz == 2){
			if(!clsBinRedundant(lit0,lit1)){
				clsCreateBinary(lit0,lit1);}
																	return true;}

		Literal lit2 = *(lbeg+2);
		if(l_sz == 3){
			if(!clsTernRedundant(lit0,lit1,lit2)){
				/*if(settings.bump_false_in_shorts){
					if(lit0.isNeg())act_hp.bump(lit0,settings.only_bump_on_card_match);
					if(lit1.isNeg())act_hp.bump(lit1,settings.only_bump_on_card_match);
					if(lit2.isNeg())act_hp.bump(lit2,settings.only_bump_on_card_match);
				}*/
				clsCreateTernary(lit0,lit1,lit2);}
																	return true;}

		assert(l_sz == 0);											return state_ok = false;
	}

	assert(l_sz <= Clause::MAX_SIZE);

	clsCreate(lbeg,lend,redundant,false);
	return true;

}


bool BasicSolver::addCardinalityConstraint(vector<Literal>& lits){

	assert(decisionLevel() == 0);

	if (!state_ok || lits.empty()){								return state_ok = false;}

	// Check if clause is satisfied and remove false/duplicate literals:
	std::sort(lits.begin(),lits.end());
	reserveForVariables(lits.back().toVar().idx()+1);

	Literal						last	(INVALID_LIT);
	vector<Literal>::iterator   to	(lits.begin());
	bool one_true = false;

	Category cat(tmp_cats_count.size());
	tmp_cats_count.push_back(0);


	for(uint i =0;i < lits.size();++i){

		if(litIsTrue(lits[i]) || lits[i] == last.flipped()){

			if(lits[i] == last.flipped()) --to; //remove last

			if(one_true) return state_ok = false;			//only one must be true

			one_true = true;

		}else if(litIsUnknown(lits[i]) && lits[i] != last){
			*to++ = last = lits[i];
			bool ok = act_hp.setCategory(last.toVar(),cat,last.pol().isTrue());
			if(!ok){print_out("Two many categories for variable %d \n",last.toVar().toDIMACS());}
			assert(ok);
		}
	}

	if(one_true){
		for(vector<Literal>::iterator iter = lits.begin(); iter != to; ++iter){
			assert(litIsUnknown(*iter));
			uncheckedEnqueue<Clause*>(iter->flipped());
		}
		return state_ok = (bcp(false) == BCP_SUCCESS);
	}

	return addValidCardConstraint(lits.begin(),to);
}




bool BasicSolver::addValidCardConstraint(vector<Literal>::iterator lbeg, const vector<Literal>::iterator lend){

	assert(decisionLevel() == 0);
	uint32_t l_sz(lend-lbeg);

	if(l_sz < 4){
		if(l_sz == 1){
			assert(litIsUnknown(*lbeg));
			uncheckedEnqueue<Clause*>(*lbeg);
			return state_ok = (bcp(false) == BCP_SUCCESS);}

		if(l_sz == 2){
			clsCreateBinary(*lbeg,*(lbeg+1));
			clsCreateBinary(lbeg->flipped(),(lbeg+1)->flipped());
			return true;
		}

		if(l_sz == 3){
			clsCreateTernary(*lbeg,*(lbeg+1),*(lbeg+2));
			clsCreateBinary(lbeg->flipped()		,(lbeg+1)->flipped());
			clsCreateBinary(lbeg->flipped()		,(lbeg+2)->flipped());
			clsCreateBinary((lbeg+1)->flipped()	,(lbeg+2)->flipped());
			return true;
		}

		assert(l_sz == 0);											return state_ok = false;
	}

	assert(l_sz <= Clause::MAX_SIZE);

	statistics.implicit_card_binaries += (l_sz *(l_sz-1))/2;

	clsCreate(lbeg,lend,false,true);
	return true;

}

/*
TrBool	BasicSolver::noicySearching	(uint count){

	TrBool status = UNKNOWN_TrBool;
	curr_rand_var_freq 	= 0.8;
	curr_rand_cat_freq	= 0.5;
	const double conflicts = curr_max_conflicts;
	curr_max_conflicts 	= 500;

	uint last = assign_stack.size();
	print_out("Random Round \n");
	while(count--){

		status		= search();

		if(console_out != NULL){ statistics.printStatsProgress(numberLiterals());}

		if(status.isKnown()) break;

		backjumpKeepLevel(0,BACKJUMP_PREV_UNTOUCHED);

		if(assign_stack.size() != last){
			print_out("Unit vars from %d to %d [rounds %d remain] ",last,assign_stack.size(),count);
			for(;last != assign_stack.size();++last){
				print_out(" %d ",assign_stack[last].toDIMACS());
			}
			print_out("\n");
		}
	}
	print_out("Finish Random Round \n");

	curr_rand_var_freq 	= settings.random_var_freq;
	curr_rand_cat_freq	= settings.random_cat_freq;
	curr_max_conflicts  = conflicts;
	curr_rand_round 	= false;
	return status;
}
*/



TrBool BasicSolver::search(){

	assert(state_ok);

	//uint64_t			max_learnts			= (uint64_t) curr_max_learnts;
	TrBool				result				= UNKNOWN_TrBool;

	uint64_t 			conflicts_remain	= curr_max_conflicts < 0 ? (~zero64): (uint64_t) curr_max_conflicts;

	//bool 				prop_from_conflict  = false;
	uint				arc_check_above 	= settings.arc_consistenct_check;

	while(conflicts_remain > 0 || decisionLevel() == 0 ){

		int32_t bcp_res	=	bcp(true);

		if(bcp_res == BCP_SUCCESS){		//there is no conflict

			if(arc_check_above < decisionLevel() || settings.randy.nextInt(1000) < settings.random_arc_check_freq_permil){
				TrBool check_arc = checkCardConstraints();
				if(check_arc.isFalse()){ return TrBool(state_ok = false);}

				if(check_arc.isUnknown()){  // not arc consistent
					++statistics.nbr_arc_const_check_worked;
					arc_check_above = decisionLevel();
					continue;
				}
				//prop_from_conflict = false;
				set_max_int(arc_check_above,decisionLevel()+1);
				++statistics.nbr_arc_const_check_failed;
				continue;
			}

			if (decisionLevel() == 0){
				result = simplify();
				if(result.isKnown()) return result;

				assert(queue_begin == assign_stack.size());

				//sumbsumption test
				if(learnts_checked_to < learnt_cls.size()){
					checkSubsumed(learnt_cls,learnts_checked_to);}
				assert(queue_begin == assign_stack.size());


				//minimising with respect to categories
				if(settings.minimise_by_cat){
					for(uint k = learnt_cls.size(); learnts_checked_to < k;){
						tryMakeShortByCardinality(*learnt_cls[--k]);
						if(!state_ok) return FALSE_TrBool;
					}
				}
				learnts_checked_to = learnt_cls.size();

				assert(queue_begin == assign_stack.size());
				//try to propagate all vars of this category
				Category cat 	= (settings.randy.nextInt(1000) < settings.random_cat_freq_permil)
								? act_hp.getRandomCategory(settings.randy) : act_hp.getMaxCategory();

				if(cat.isValid() && firstDecisionCheckAll(cat)) continue;



			}else if(decisionLevel() == 1){
				uint start = decision_level_end[0];
				assert(assign_stack.size() > start);
				const Literal flp_dec_lit = assign_stack[start].flipped();
				while(++start < assign_stack.size()){
					Literal impl = assign_stack[start];
					if(!clsBinRedundant(flp_dec_lit,impl)){
						clsCreateBinary(flp_dec_lit,impl);
					}
				}
			}else if(decisionLevel() == 2){
				uint start1 = decision_level_end[0];
				uint start2 = decision_level_end[1];
				assert(assign_stack.size() > start2);
				const Literal flp_dec_lit1 = assign_stack[start1].flipped();
				const Literal flp_dec_lit2 = assign_stack[start2].flipped();
				while(++start2 < assign_stack.size()){
					Literal impl = assign_stack[start2];
					if(!clsTernRedundant(flp_dec_lit1,flp_dec_lit2,impl)){
						clsCreateTernary(flp_dec_lit1,flp_dec_lit2, impl);
						//print_out("Added clause at level 1 %d %d \n",flp_dec_lit.toDIMACS(),impl.toDIMACS());
					}
				}
			}

			/*uint64_t learnts_allowed = max_learnts +numberAssignments();
			if(learnts_allowed <= numberLearntLongClses()){learntsGC((learnts_allowed * 3)/4); }*/


			Literal next	= nextDecision(settings.polarity_mode);
			if (next == INVALID_LIT){															return TRUE_TrBool;   }

			// Increase decision level and enqueue 'next'
			assert(litIsUnknown(next));
			newDecisionLevel();
			uncheckedEnqueue<Clause*>(next);
			//prop_from_conflict = false;
/*
#ifndef USE_CATEGORY
			if(decisionLevel() == 1){firstDecisionCheckBothPol(next);}
#endif
*/
			//print_out("D: %d @ %d\n",next.toDIMACS(), decisionLevel());

			set_max_int(statistics.max_dec_level,decisionLevel());


			//next = INVALID_LIT;

		}else{						//there is a conflict

			++statistics.nbr_conflicts;
			--conflicts_remain;
			//prop_from_conflict = true;

			if (!firstUIP(bcp_res))  {																return TrBool(state_ok = false); }
			assert(litAssignLevel(conflict_lemma[0]) == decisionLevel());

			reduceSelfImpl();

			uint32_t backtrack_level = backtrackLevel();

			int32_t bj_dist = decisionLevel() - backtrack_level;
			statistics.storeBackjumpingDistance(bj_dist);

			backjumpKeepLevel(backtrack_level,BACKJUMP_SET_PREV);
			assert(litIsUnknown(conflict_lemma[0]));

			const uint32_t l_sz(conflict_lemma.size());
			if(l_sz > 3){
				assert(l_sz <= Clause::MAX_SIZE);
				Clause* learnt_cl = clsCreate(conflict_lemma.begin(),conflict_lemma.end(),true,false);
				if(learnt_cl == NULL){break;}
				clsBump(learnt_cl);
				uncheckedEnqueue(conflict_lemma[0], learnt_cl);
			}else if(l_sz == 3){
				assert(!clsTernRedundant(conflict_lemma[0],conflict_lemma[1],conflict_lemma[2]));
				clsCreateTernary(conflict_lemma[0],conflict_lemma[1],conflict_lemma[2]);
				uncheckedEnqueue(conflict_lemma[0],conflict_lemma[1],conflict_lemma[2]);
			}else if(l_sz == 2){
				assert(!clsBinRedundant(conflict_lemma[0],conflict_lemma[1]));
				clsCreateBinary(conflict_lemma[0],conflict_lemma[1]);
				uncheckedEnqueue(conflict_lemma[0],conflict_lemma[1]);
			}else{
				clsLearnUnit(conflict_lemma[0]);
			}

			act_hp.decayAll();
			claDecayActivity();

			if ( numberLiterals() > settings.max_literals){
				print_out("Force Learnts GC due to Memory: %llu Literals, and %u allowed\n",numberLiterals(),settings.max_literals);
				learntsGC(numberLearntLongClses() / 3);
				++statistics.nbr_panic_gc;
				//backjumpKeepLevel(0,BACKJUMP_SET_PREV_FLP);
/*
				if(settings.dimacs_file_out != NULL){
					char buf[strlen(settings.dimacs_file_out)+10];
					sprintf(buf,"%s_gc_known%d.cnf",settings.dimacs_file_out,statistics.nbr_panic_gc);
					OFStream out(buf,true);
					printDIMACS(out,true,false,false,false);
				}
*/
				learnts_checked_to = numberLearntLongClses();
				reduced_learnts = true;

				//break;
			}
		}
	}

	//maximal number of conflicts reached
	return UNKNOWN_TrBool;

}


void	BasicSolver::backjumpUndoAssigns (uint32_t assign_idx,BackjumpType type){

	Variable x = INVALID_VAR;
	assert(assign_stack.size() >= assign_idx);

	queue_begin 	= assign_stack.size();

	if(type == BACKJUMP_SET_PREV && settings.polarity_mode != Polarity_Best){
		while(queue_begin != assign_idx ){
			x  = assign_stack[--queue_begin].toVar();
			varClearAssignmentSetPrev(x);
			reinsertVarOrder(x);}
	}else if(type == BACKJUMP_SET_PREV_FLP && settings.polarity_mode != Polarity_Best){
		while(queue_begin != assign_idx ){
			x  = assign_stack[--queue_begin].toVar();
			varClearAssignmentSetPrevFlp(x);
			reinsertVarOrder(x);}
	/*}else if(type == BACKJUMP_SET_PREV_RAND){
		while(queue_begin != assign_idx ){
			x  = assign_stack[--queue_begin].toVar();
			settings.randy.nextBool() ? varClearAssignmentSetPrevFlp(x):varClearAssignmentSetPrev(x) ;
			reinsertVarOrder(x);}*/
	}else{
		while(queue_begin != assign_idx ){
			x  = assign_stack[--queue_begin].toVar();
			varClearAssignmentSimple(x);
			reinsertVarOrder(x);}
	}

	assign_stack.resize(queue_begin);

/*
	if(++clause_stamp > Clause::MAX_STAMP){
		clsClearAllStamps();
		clause_stamp = 0;
		++statistics.nbr_expensive_cls_stamp;
	}
*/
}

bool BasicSolver::clsImpliedByCardConstraint(const Literal lit0, const Literal lit1){
	assert(lit0 != INVALID_LIT && lit1 != INVALID_LIT);
	//if(statistics.parse_timestmp != 0.0) print_out("Checking pair %d %d\n",lit0.toDIMACS(),lit1.toDIMACS());
	const CategoryPair& c0p = act_hp.getCats(lit0.toVar());
	const CategoryPair& c1p = act_hp.getCats(lit1.toVar());

	Category c0_1 = c0p.getFirst();	Category c0_2 = c0p.getSecond();
	Category c1_1 = c1p.getFirst();	Category c1_2 = c1p.getSecond();

	return
	//bool ret=
	(
		(
			c0_1.isValid() && (c0p.polarity(true) == lit0.isNeg()) //lit0 is opposite to the polarity of its first category
		)
		&&
		(
				(c1_1==c0_1 && c1p.polarity(true)==lit1.isNeg())    //first category of lit1 equals first category of lit0 and also opposite polarity
			||	(c1_2==c0_1 && c1p.polarity(false)==lit1.isNeg())  //second cat of lit1 ...
		)
	)
	||
	(
		(
			c0_2.isValid() && (c0p.polarity(false) == lit0.isNeg()) //lit0 is opposite to the polarity of its second category
		)
		&&
		(
				(c1_1==c0_2 && c1p.polarity(true)==lit1.isNeg())    //first category of lit1 equals second category of lit0 and also opposite polarity
			||	(c1_2==c0_2 && c1p.polarity(false)==lit1.isNeg()) //second cat of lit1 ...
		)
	);

	/*if(ret){
		print_out("Bin implied by Cat %d %d \n", lit0.toDIMACS(), lit1.toDIMACS());
	}

	return ret;*/
}


Variable BasicSolver::nextNoCatDecision(){

	const bool rand	= settings.randy.nextInt(1000) < settings.random_var_freq_permil;
	if(rand)++statistics.nbr_rnd_decisions;else ++statistics.nbr_act_decisions;
	Variable var = INVALID_VAR;

	while(!act_hp.emptyDefaultCat()){
		var = rand 	?	act_hp.removeRandomDefaultCat(settings.randy)
					:	act_hp.removeMaxDefaultCat();
		assert(var.idx() < numberVariables());
		assert(var.idx() >= 0);
		assert(var != INVALID_VAR);
		if(varValue(var).isUnknown()) return var;
	}

	return INVALID_VAR;
}

Variable BasicSolver::nextDecisionVar(const Category& cat){

	assert(cat.isValid());

	const bool rand	= settings.randy.nextInt(1000) < settings.random_var_freq_permil;
	if(rand)++statistics.nbr_rnd_decisions;else ++statistics.nbr_act_decisions;
	Variable 	var = INVALID_VAR;
	TrBool 		val;

	while(!act_hp.empty(cat)){

		var	= rand 	? 	act_hp.removeRandom(cat,settings.randy)
					:	act_hp.removeMax(cat);

		if(var == INVALID_VAR || (val = varValue(var),val.isUnknown())) return var;

		/*not problematic since everything is propagated*/
		if(act_hp.varPolarityAtCat(var,cat) == val.isTrue()){
			act_hp.removeAll(cat); break;}
	}

	return INVALID_VAR;
}


Literal	BasicSolver::decisionLitPolarity(Variable var,PolarityMode mode, Category cat ){
	assert(var!= INVALID_VAR);

	switch (mode){
		case Polarity_False	: return var.toNegLit();
		case Polarity_True	: return var.toPosLit();
		case Polarity_Rand	: return settings.randy.nextBool() ?
									var.toNegLit():var.toPosLit();
		case Polarity_Prev	: return varPrevValue(var).isFalse() ?
									var.toNegLit() : var.toPosLit();
		case Polarity_Card	: return (cat.isValid() && act_hp.varPolarityAtCat(var,cat)) ?
									var.toPosLit() : var.toNegLit();
		case Polarity_Best	: return varPrevValue(var).isFalse() ?
									var.toNegLit() : var.toPosLit();
		default: assert(false); return var.toNegLit();
	}
}


Literal BasicSolver::nextDecision(PolarityMode mode){

	const bool rand = settings.randy.nextInt(1000) < settings.random_cat_freq_permil;
	if(rand)++statistics.nbr_rnd_decisions; else ++statistics.nbr_act_decisions;

	Category cat; assert(!cat.isValid());
	Variable var = INVALID_VAR;

	do{
		if(act_hp.empty()) 			return INVALID_LIT;

		if(settings.polarity_mode == Polarity_Card){
			cat = rand 	? 	act_hp.removeRandomCategory(settings.randy):
							act_hp.removeMaxCategory();

		}else{

			if(cat.isValid()){act_hp.removeCat(cat);} //previous round did not get variable out of this category

			cat = rand 	? 	act_hp.getRandomCategory(settings.randy):
							act_hp.getMaxCategory();
		}

		var = cat.isValid() ? nextDecisionVar(cat) : nextNoCatDecision();

	}while(var == INVALID_VAR);


	assert(var != INVALID_VAR );

	if(settings.polarity_mode == Polarity_Card){
		return decisionLitPolarity(var,mode,cat);
	}

	Literal lit = decisionLitPolarity(var,mode,cat);
	if(cat.isValid() && act_hp.varPolarityAtCat(var,cat) == lit.isPos()){
		act_hp.removeCat(cat);
	}

	return lit;
}



bool	BasicSolver::firstDecisionCheckAll(Category cat){
	assert(decisionLevel() == 0);
	assert(cat.isValid());
	assert(tmp_literal_store.empty());
	assert(queue_begin == assign_stack.size());

	const uint32_t assign_start = assign_stack.size();
	const uint32_t impl_start = assign_start+1;


	Literal next;
	TrBool val;
	int 	tmp_lits_size = -1;
	//get information of implication of first choice
	int k = act_hp.size(cat);
	while(k--){
		Variable curr_var = act_hp.get(cat,k);
		assert(curr_var != INVALID_VAR);

		if((val = varValue(curr_var),val.isKnown())){
			if(act_hp.varPolarityAtCat(curr_var,cat) == val.isTrue()){
				assert(tmp_lits_size < 0 && tmp_literal_store.empty());
				return false;
			}
			continue;
		}


		next = act_hp.varPolarityAtCat(curr_var,cat) ? curr_var.toPosLit() : curr_var.toNegLit();
		uncheckedEnqueue<Clause*>(next);

		if(BCP_SUCCESS == bcp(false)){

			if(tmp_lits_size >= 0){
				assert(tmp_lits_size != 0);

				uint32_t to = 0;
				for(int32_t i = 0; i != tmp_lits_size;++i){
					Literal lit = tmp_literal_store[i];
					if(litIsTrue(lit)){tmp_literal_store[to++] = lit;}
				}
				tmp_lits_size = to;

			}else{

				for(uint32_t i = impl_start; i != assign_stack.size();++i){
					tmp_literal_store.push_back(assign_stack[i]);};
				tmp_lits_size = tmp_literal_store.size();
			}

			backjumpUndoAssigns(assign_start,BACKJUMP_PREV_UNTOUCHED);
			if(tmp_lits_size == 0) {
				tmp_literal_store.clear();
				return false;
			}

		}else{ //conflict at first decision -> opposite is implied
			backjumpUndoAssigns(assign_start,BACKJUMP_PREV_UNTOUCHED);
			uncheckedEnqueue<Clause*>(next.flipped());
			tmp_literal_store.clear();
			return true;
		}
	}

	assert(tmp_lits_size != 0 && assign_stack.size() == assign_start);

	if(tmp_lits_size > 0) {
		statistics.nbr_flipped_prop_learnts += tmp_lits_size;
		while(tmp_lits_size > 0){uncheckedEnqueue<Clause*>(tmp_literal_store[--tmp_lits_size]);}
		tmp_literal_store.clear();
	}

	return tmp_lits_size >= 0;
}



TrBool	BasicSolver::preProcess	(){

	//uint unit_clses 		= assign_stack.size();
	//print_out("Unit Clauses %d \n",unit_clses);

	TrBool status = UNKNOWN_TrBool;
	curr_max_conflicts = settings.preprocess_conflicts;
	vector<Category> cats_to_do;

	for(uint i = numberCats(); i--;){ cats_to_do.push_back(Category(i));}

	if(settings.preproc_shuffle){
		std::random_shuffle(cats_to_do.begin(),cats_to_do.end(),settings.randy);}

	uint bin_clses = numberBinClses() + settings.bin_cls_to_force_closure;

	while(!cats_to_do.empty()){
		Category cat = cats_to_do.back(); cats_to_do.pop_back();
		assert(decisionLevel() == 0);

		if(firstDecisionCheckAll(cat)){
			state_ok = (bcp(false) == BCP_SUCCESS);
			if(!state_ok) return FALSE_TrBool;
		}

		vector<Variable> vars_of_cat;

		if(settings.preprocess_level > 1){
			for(uint k = act_hp.size(cat);k--;){vars_of_cat.push_back(act_hp.get(cat,k));}
			if(settings.preproc_shuffle){
				std::random_shuffle(vars_of_cat.begin(),vars_of_cat.end(),settings.randy);}
		}else{
			vars_of_cat.push_back(nextDecisionVar(cat));
		}

		while(!vars_of_cat.empty()){
			Variable var = vars_of_cat.back(); vars_of_cat.pop_back();
			if(var == INVALID_VAR || varValue(var).isKnown()) continue;

			Literal next = decisionLitPolarity(var,settings.polarity_mode,cat);

			//if(settings.flip_first_in_preproc){	next = next.flipped();}

			newDecisionLevel();
			uncheckedEnqueue<Clause*>(next);

			status = search();

			if(console_out != NULL){ statistics.printStatsProgress(console_out, numberLiterals());}

			if(status.isKnown()) {return status;}

			backjumpKeepLevel(0,BACKJUMP_SET_PREV);

			if(reduced_learnts || bin_clses < numberBinClses()){

				bin_clses = numberBinClses() + settings.bin_cls_to_force_closure;;

				console_out->writeString("Compute Closure \n");
				status= clsComputeBinTernClosure();
				if(status.isKnown()) break;

				console_out->writeString("Clean fix clauses. %d rounds remain in preprocess. \n",cats_to_do.size());
				status = checkSubsumed(fix_cls);
				if(status.isKnown()) break;

				console_out->writeString("Clean learnts \n");
				status = checkSubsumed(learnt_cls);
				if(status.isKnown()) break;

				reduced_learnts = false;
			}

			if(learnts_checked_to < learnt_cls.size()){
				checkSubsumed(learnt_cls,learnts_checked_to);}

			if(settings.minimise_by_cat){
				for(uint k = learnt_cls.size(); learnts_checked_to < k;){
					tryMakeShortByCardinality(*learnt_cls[--k]);
					if(!state_ok) return FALSE_TrBool;
				}
			}
			learnts_checked_to = learnt_cls.size();
		}
	}

	curr_max_conflicts = settings.conflict_max_init;

	return status;
}

bool BasicSolver::debugTryMakeShortPossible(Clause& cl, Literal lit0, Literal lit1, Literal lit2){
	assert(lit0 != INVALID_LIT);
	assert(decisionLevel() == 0);
	for(uint i = cl.size(); i--;){
		Literal l = cl[i];
		if(litIsTrue(l)) break;
		if(litIsFalse(l) || l == lit0 || l == lit1 || l == lit2) continue;

		Literal f = l.flipped();
		if(clsBinRedundant(f,lit0)) continue;
		if(lit1 != INVALID_LIT && clsBinRedundant(f,lit1)) continue;
		if(lit2 != INVALID_LIT && clsBinRedundant(f,lit2)) continue;

		console_out->writeString("Bad things happen: %d ", lit0.toDIMACS());
		if(lit1 != INVALID_LIT) console_out->writeString("%d ", lit1.toDIMACS());
		if(lit2 != INVALID_LIT) console_out->writeString("%d ", lit2.toDIMACS());
		console_out->writeString("cannot be resolved from: \n");
		printClause(*console_out,cl);
		return false;
	}


	return true;
}



void BasicSolver::tryMakeShortByCardinality(Clause& cl){

	assert(allEqual<int16_t>(tmp_cats_count,0));
	assert(!cl.isCardConstraint());
	assert(var_flags.allCleared());
	assert(queue_begin == assign_stack.size());

	Literal new_cls[3];										uint n_sz=0;
	pair<CategoryPair,Literal>  choice_lits[cl.size()]; 	uint c_sz=0;
	Category touched_cats[cl.size()*2 +3 ]; 				uint t_sz=0;

	/*OFStream out(stderr);
	out.writeString("Check clause \n");*/
	//printClause(out,cl);

	//seperate in fixed lits and those that might be deleted
	for(uint i=0; i < cl.size();++i){
		const Literal lit 	= cl[i];
		if(litIsTrue(lit)){	goto FINISH_THIS_CLS;}
		if(litIsFalse(lit)) continue;

		const bool pol		= lit.isPos();
		const Variable var 	= lit.toVar();
		var_flags.setFlag(var);

		const CategoryPair& cats = act_hp.getCats(var);
		//out.writeString("Var %d has cat %d and %d \n", var.toDIMACS(), cats.getFirst().idx(), cats.getSecond().idx());

		if(cats.hasTwoCategories() && (cats.polarity(true)==pol) && (cats.polarity(false)==pol)){
			int16_t& cat1_count	= tmp_cats_count[cats.getFirst().idx()] ;
			int16_t& cat2_count	= tmp_cats_count[cats.getSecond().idx()];
			touched_cats[t_sz++] = cats.getFirst();
			touched_cats[t_sz++] = cats.getSecond();

			if(cat1_count >=0 && cat2_count >=0){
				++cat1_count; ++cat2_count;
				choice_lits[c_sz++] = pair<CategoryPair,Literal>(cats,lit);
			}//else lit can be omitted since binary clause exists that allows resolution over lit to a literal already covered
		}else{

			if(litIsUnknown(lit)) {
				if(n_sz >= 3){goto FINISH_THIS_CLS;}
				new_cls[n_sz++] = lit;
			}
			if(cats.categoriesed() && cats.polarity(true)!=pol){
				tmp_cats_count[cats.getFirst().idx()] = -1; touched_cats[t_sz++] = cats.getFirst();}

			if(cats.hasTwoCategories() && cats.polarity(false)!=pol){
				tmp_cats_count[cats.getSecond().idx()] = -1;touched_cats[t_sz++] = cats.getSecond();}
		}
	}

	while(c_sz ){ //while there are choice variables

		//filter
		for(uint i=0;i<c_sz;){
			CategoryPair& cats = choice_lits[i].first;
			int16_t& cat1_count	= tmp_cats_count[cats.getFirst().idx()] ;
			int16_t& cat2_count	= tmp_cats_count[cats.getSecond().idx()];
			bool del = false;
			assert(cat1_count != 0 && cat2_count != 0);
			/*if(cat1_count == 1 && cat2_count == 1){//no other literal has one category in common
				cat1_count = -1; cat2_count = -1;
				Literal& lit = choice_lits[i].second;
				if(n_sz >= 3){goto FINISH_THIS_CLS;}
				new_cls[n_sz++] = lit; //keep this literal in clause
				del = true;
			}else*/ if(cat1_count < 0 ){ del=true; --cat2_count; //already covered by another
			}else if(cat2_count < 0 ){ del=true; --cat1_count;}

			if(del)	{	choice_lits[i]=choice_lits[--c_sz]; }
			else	{	++i;}
		}

		if(c_sz == 0) break;
		if(n_sz >= 3){goto FINISH_THIS_CLS;}

		//find max category
		Category max; int16_t max_occur = -1;
		for(uint i= 0;i<t_sz;){
			int16_t& count = tmp_cats_count[touched_cats[i].idx()];
			if(count <= 0){
				count = 0;
				touched_cats[i] = touched_cats[--t_sz];
				continue;
			}

			if(count > max_occur){	max_occur = count;	max = touched_cats[i];}
			else if(count == max_occur){max = act_hp.determineMax(max,touched_cats[i]);}
			++i;
		}


		//choose one of category
		if(!max.isValid()) {assert(false);goto FINISH_THIS_CLS;}

		for(uint i=0;i<act_hp.size(max) ;++i){

			Variable var = act_hp.get(max,i);
			if(!var_flags.isSet(var)){
				if(varValue(var).isKnown()) continue;
				Literal lit = act_hp.varPolarityAtCat(var,max) ? var.toNegLit() : var.toPosLit();
				assert(n_sz < 3);
				new_cls[n_sz++] = lit;

				tmp_cats_count[max.idx()] = -1;
				const CategoryPair& cats = act_hp.getCats(var);
				assert(cats.getFirst() == max || cats.getSecond() == max);
				if(cats.hasTwoCategories()){
					Category other_cat = cats.getFirst() == max ? cats.getSecond() : cats.getFirst();
					assert(other_cat.isValid());
					if(act_hp.varPolarityAtCat(var,other_cat) == lit.isNeg()){
						tmp_cats_count[other_cat.idx()] = -1;
						assert(t_sz < cl.size()*2 + 3);
						touched_cats[t_sz++] = other_cat;
					}else{
						var_flags.setFlag(var);
					}
				}
				break;
			}
		}

		if(tmp_cats_count[max.idx()] != -1){ goto FINISH_THIS_CLS;}
	}


	assert(n_sz <= 3);

	if(n_sz == 3){
		assert(new_cls[0] != new_cls[1]);
		assert(new_cls[0] != new_cls[2]);
		assert(new_cls[1] != new_cls[2]);
		if(clsTernRedundant(new_cls[0],new_cls[1],new_cls[2])){
			++statistics.nbr_cat_minimize_fail_redundant;
		}else{
			++statistics.nbr_cat_minimize_worked;
			statistics.nbr_cat_minimize_lits += (cl.size() - n_sz);
			clsCreateTernary(new_cls[0],new_cls[1],new_cls[2]);
			assert(debugTryMakeShortPossible(cl,new_cls[0],new_cls[1],new_cls[2]));
		}
	}else if(n_sz == 2){
		assert(new_cls[0] != new_cls[1]);
		if(clsBinRedundant(new_cls[0],new_cls[1])){
			++statistics.nbr_cat_minimize_fail_redundant;
		}else{
			++statistics.nbr_cat_minimize_worked;
			statistics.nbr_cat_minimize_lits += (cl.size() - n_sz);
			clsCreateBinary(new_cls[0],new_cls[1]);
			assert(debugTryMakeShortPossible(cl,new_cls[0],new_cls[1]));
		}
	}else{
		++statistics.nbr_cat_minimize_worked;
		statistics.nbr_cat_minimize_lits += (cl.size() - n_sz);
		assert(n_sz == 1);
		clsLearnUnit(new_cls[0]);
		assert(debugTryMakeShortPossible(cl,new_cls[0]));
		state_ok = (bcp(false) == BCP_SUCCESS);
	}


	FINISH_THIS_CLS: ;
	for(uint i=cl.size();i--;) {var_flags.carelessClear(cl[i].toVar());}
	while(n_sz--) {var_flags.carelessClear(new_cls[n_sz].toVar());}
	while(t_sz--){tmp_cats_count[touched_cats[t_sz].idx()] = 0;}
	assert(allEqual<int16_t>(tmp_cats_count,0));
	assert(var_flags.allCleared());
}










int32_t	BasicSolver::bcp(const bool init_conflict){

	int32_t			nbr_pivot_lits = BCP_SUCCESS; //indicates no conflict by now
	uint32_t		num_props(0);

	while(queue_begin < assign_stack.size()){

		++num_props;
		nbr_pivot_lits = bcpPropagateLit(init_conflict,assign_stack[queue_begin++].flipped());
	}

	statistics.nbr_propagations	+= num_props;
	simplify_props_remaining		-= num_props;

	return nbr_pivot_lits;
}


int32_t	BasicSolver::bcpPropagateLit(const bool init_conflict, const Literal& false_lit){

	we_vec& watched	= litWatchedClauses(false_lit); //watched clauses that have to be checked when lit gets false
	we_iter iter = watched.begin();
	we_iter copy = iter;
	we_iter end  = watched.end();

	while(iter != end){
		WatchedElement& we = *iter++;

		if(we.isCardinality()){
			*copy++ = we; //keep this watcher
			Clause& card_cls = *we.getClause();
			assert(card_cls.isCardConstraint());

			for(uint k= card_cls.size();k--;){
				Literal impl = card_cls[k].flipped();
				if(impl==false_lit) continue; //card_cls[k] is the one to be true

				//implied binary clause
				if(litIsUnknown(impl)){
					uncheckedEnqueue<const Literal&>(impl,false_lit);
				}else if(litIsFalse(impl)){
					bcpFinishOnConflict(watched,iter,copy,end);
					return init_conflict ? bcpInitConflict(decisionLevel(),false_lit,impl):1;
				}
			}
			continue;
		}


		//might be any other literal of this clause except the watched literal
		const Literal other = we.getFirstLiteral();
		assert(other != false_lit);

		//keep this watcher if other lit is true
		if(litIsTrue(other)){ *copy++ = we; continue;}

		if(we.isBinary()){
			*copy++ = we; //keep this watcher
			if(litIsUnknown(other)){
				uncheckedEnqueue<const Literal&>(other,false_lit);
				continue;
			}else{ //other literal in binary clause is false --> conflict
				bcpFinishOnConflict(watched,iter,copy,end);
				return init_conflict ? bcpInitConflict(decisionLevel(),false_lit,other):1;
			}
		}


		if(we.isTernary()){
			*copy++ = we; //keep this watcher
			const Literal third = we.getSecondLiteral();
			if(litIsTrue(third)){
				we.setFirstLiteral(third);
				we.setSecondLiteral(other);
				continue;
			}

			if(litIsUnknown(other)){
				if(litIsFalse(third)){
					uncheckedEnqueue(other,false_lit,third);
				}
				continue;
			}else{
				assert(litIsFalse(other));
				if (litIsFalse(third)){
					bcpFinishOnConflict(watched,iter,copy,end);

					if(init_conflict){
						int32_t  nbr_pivot_lits = bcpInitConflict(decisionLevel(),false_lit,other);
						firstUIPCheckLiteral(third,decisionLevel(),nbr_pivot_lits);
						assert(nbr_pivot_lits != 1);
						return nbr_pivot_lits;
					}else{
						return 1; //any pos value
					}
				}
			}
			assert(litIsFalse(other) && litIsUnknown(third));
			uncheckedEnqueue(third,false_lit,other);
			continue;
		}

		Clause& cl = *(we.getClause());

		uint32_t other_watcher_index = (cl[1] != false_lit);
		assert(cl[other_watcher_index ^ 1] == false_lit && cl[other_watcher_index] != false_lit);

		//it might be that the shortcut literal in 'we' does not point to the actual watched literal anymore
		if(litIsTrue(cl[other_watcher_index])){
			assert(!we.isShort());
			we.setFirstLiteral(cl[other_watcher_index]);
			*copy++ = we; //keep this watcher
			continue;
		}

		if(bcpFindNewWatcher(cl,false_lit,other_watcher_index)){continue;}


		// No new watched found -- clause is unit under assignment:
		we.setFirstLiteral(cl[other_watcher_index]);
		*copy++				= we;

		if (litIsFalse(cl[other_watcher_index])){
			if(cl.isLearnt())clsBump(&cl);

			bcpFinishOnConflict(watched,iter,copy,end);

			if(init_conflict){
				const int32_t dec_level = decisionLevel();
				int32_t  nbr_pivot_lits = bcpInitConflict(dec_level,false_lit,cl[other_watcher_index]);
				for(uint32_t k = 2;k<cl.size();++k){firstUIPCheckLiteral(cl[k],dec_level,nbr_pivot_lits);}
				assert(nbr_pivot_lits != 1);
				return nbr_pivot_lits;
			}else{
				return 1; //any positive value
			}

		}else{
			uncheckedEnqueue(cl[other_watcher_index], &cl);
		}

	}
	if(copy != end) watched.erase(copy,end);

	return BCP_SUCCESS;
}




bool BasicSolver::firstUIP(int32_t nbr_pivot_elements){
	assert(nbr_pivot_elements >= 0);
	if(nbr_pivot_elements == 0) return false;

	const uint32_t   dec_level		(decisionLevel());
	uint32_t		assign_stack_index  = assign_stack.size() ;

	assert(nbr_pivot_elements != 1);
	AssignInfo 	confl_info;
	Literal 	pivot_lit;

	while(true){

		//skip all unseen assignments
		//since there always exists a UIP the index cannot become negative
		while (!var_flags.isSet(assign_stack[--assign_stack_index].toVar())){}
		assert(assign_stack_index >= 0);

		pivot_lit					= assign_stack[assign_stack_index];
		//can be already cleared since this lit cannot appear in any remaining (previous) reason because literals
		//in reason clauses are all false (except for the one it is the assign reason)
		var_flags					.clear(pivot_lit.toVar());

		if(--nbr_pivot_elements == 0){
			assert(litAssignLevel(pivot_lit) == dec_level);
			assert(pivot_lit != INVALID_LIT);
			conflict_lemma[0]				= pivot_lit.flipped();
			return true;
		}

		confl_info					= varAssignInfo(pivot_lit.toVar());
		assert(litAssignLevel(pivot_lit) == dec_level);

		assert(!confl_info.noReason());

		if(confl_info.isShortReason()){
			firstUIPCheckLiteral(confl_info.getShortFirstReason(),dec_level,nbr_pivot_elements);

			if(confl_info.isTernReason()){
				firstUIPCheckLiteral(confl_info.getShortSecondReason(),dec_level,nbr_pivot_elements);
			}

		}else{
			assert(confl_info.isClsReason() && confl_info.getClsReason() != NULL);

			Clause& cl					= *confl_info.getClsReason();
			const uint32_t cl_sz			(cl.size());

			assert(pivot_lit == cl[0]|| pivot_lit == cl[1]);
			assert(litAssignLevel(cl[0]) == dec_level && litAssignLevel(cl[1]) == dec_level);

			if(cl.isLearnt())				clsBump(&cl);

			uint32_t cl_index = pivot_lit == cl[0];//pivot_lit ==cl[0] ? 1 : 0
			assert(cl[cl_index ^1] == pivot_lit);

			//do for all except the pivot_lit
			firstUIPCheckLiteral(cl[cl_index],dec_level,nbr_pivot_elements);
			for(cl_index = 2;cl_index < cl_sz;++cl_index){
				firstUIPCheckLiteral(cl[cl_index],dec_level,nbr_pivot_elements); }
		}
	}
	return false;
}

uint32_t BasicSolver::backtrackLevel(){
	assert(litAssignLevel(conflict_lemma[0]) == decisionLevel());
	uint32_t lsz = conflict_lemma.size();
	if(lsz < 2) return 0;

	uint32_t max_idx = 1;
	while(--lsz > 1){
		if(litAssignLevel(conflict_lemma[lsz]) > litAssignLevel(conflict_lemma[max_idx])){max_idx = lsz;}}

	//such that the literals with maximum level are the watchers
	//this is important for backtracking!
	Literal highest				= conflict_lemma[max_idx];
	conflict_lemma[max_idx]	= conflict_lemma[1];
	conflict_lemma[1]		= highest;


	return litAssignLevel(conflict_lemma[1]);
}


void	BasicSolver::reduceSelfImpl(){

	if(conflict_lemma.size() < 2){assert(var_flags.allCleared()); return;}

	//sort(conflict_lemma.begin()+1,conflict_lemma.end(),LitActDescendingFunctor(&hp_act_mp));
	assert(var_flags_to_clear.empty());

	uint cp,idx;
	for(cp = idx = 1; idx != conflict_lemma.size(); ++idx){
		if(litIsNecessary(conflict_lemma[idx])) {conflict_lemma[cp++] = conflict_lemma[idx];}}

	conflict_lemma.resize(cp);

	unsetConfLemmaFlags();
}


bool	BasicSolver::litIsNecessary(Literal lit){

	if(varAssignInfo(lit.toVar()).noReason()) return true;

	const Variable start(lit.toVar());

	uint32_t unmark_top(var_flags_to_clear.size());
	assert(tmp_literal_store.empty());

	while(true){

		assert(!varAssignInfo(lit.toVar()).noReason());
		const AssignInfo& a_info = varAssignInfo(lit.toVar());

		bool is_nec;

		if(a_info.isShortReason()){
			is_nec = reduceSelfImpCheckLiteral(a_info.getShortFirstReason());

			if(!is_nec && a_info.isTernReason()){
				is_nec = reduceSelfImpCheckLiteral(a_info.getShortSecondReason());
			}
		}else{
			assert(a_info.isClsReason() && a_info.getClsReason() != NULL);

			Clause& cl	= *(a_info.getClsReason());

			uint32_t cl_index = lit.toVar() == cl[0].toVar();
			assert(cl[cl_index^1].toVar() == lit.toVar());
			assert(var_flags.isSet(cl[cl_index^1].toVar()));

			is_nec = reduceSelfImpCheckLiteral(cl[cl_index]);
			for(cl_index = 2;!is_nec && cl_index < cl.size();++cl_index){
				is_nec = reduceSelfImpCheckLiteral(cl[cl_index]);
			}
		}

		if(is_nec){
			for( ;unmark_top < var_flags_to_clear.size();++unmark_top)   {var_flags.clear(var_flags_to_clear[unmark_top]);}
			for(int i = tmp_literal_store.size();i--;)				{var_flags.clear(tmp_literal_store[i].toVar());}
			tmp_literal_store.clear();
			return true;
		}


		if(tmp_literal_store.empty()){
			statistics.total_lemma_shortening += 1;
			var_flags_to_clear.push_back(start);
			return false;
		}

		lit = tmp_literal_store.back(); tmp_literal_store.pop_back();
		var_flags_to_clear.push_back(lit.toVar());
	}
}


void BasicSolver::clearShortWatchers(Literal lit){

	assert(litIsUnknown(lit));
	assert(queue_begin == assign_stack.size());

	we_iter from,to,end;
	we_vec& wes = litWatchedClauses(lit);

	from = to = wes.begin(); end = wes.end();
	while(from != end){
	WatchedElement& we = *from++;
	if(we.isShort()){

			if(	litIsTrue(we.getFirstLiteral()) ||
				(we.isTernary() && litIsTrue(we.getSecondLiteral()))){continue;} //delete watcher

			if(litIsFalse(we.getFirstLiteral())){
				assert(we.isTernary());
				assert(litIsUnknown(we.getSecondLiteral()));
				*to++ = WatchedElement(we.getSecondLiteral());
				continue;}

			if(we.isTernary() && litIsFalse(we.getSecondLiteral())){
				assert(litIsUnknown(we.getFirstLiteral()));
				*to++ = WatchedElement(we.getFirstLiteral());
				continue;}
		}

		*to++ = we;
	}
	if(to != end){wes.erase(to,end);}
}



void BasicSolver::clearShortWatchers(Variable var){
	assert(var.idx() < numberVariables());
	assert(var.idx() >= 0);
	if(var_flags.isSet(var) || varValue(var).isKnown()) return;
	var_flags.setFlag(var);	var_flags_to_clear.push_back(var);

	clearShortWatchers(var.toNegLit());
	clearShortWatchers(var.toPosLit());
}

TrBool BasicSolver::simplify( ){

	assert(decisionLevel() == 0);

	if (!state_ok || bcp(false) != BCP_SUCCESS){   return TrBool(state_ok = false);}

	if (assign_stack.size() != simplify_assign_idx  && (simplify_props_remaining <= 0)){

		++statistics.nbr_simplify_execs;

		removeSat(learnt_cls);
		removeSat(fix_cls);

		simplifyAllShortWatchers();

		simplify_props_remaining   = numberLiterals();
	}

	return UNKNOWN_TrBool;
}


void	BasicSolver::simplifyShortWatchersOf (we_vec& wes){
	we_iter from,to;

	for(from = to = wes.begin(); from != wes.end();++from){
		if(from->isShort()){
			clearShortWatchers(from->getFirstLiteral().toVar());
			if(from->isTernary()){clearShortWatchers(from->getSecondLiteral().toVar());};
		}else{
			//might be since bigsimplify changes are not simplified further
			*to++ = *from;
		}
	}
	wes.erase(to,wes.end());
}



void	BasicSolver::simplifyAllShortWatchers (){
	assert(var_flags.allCleared());
	assert(decisionLevel() == 0);

	while(simplify_assign_idx < assign_stack.size()){
		Literal lit = assign_stack[simplify_assign_idx++];
		simplifyShortWatchersOf(litWatchedClauses(lit));
		simplifyShortWatchersOf(litWatchedClauses(lit.flipped()));
	}


	for(uint32_t i = var_flags_to_clear.size();i--;){var_flags.carelessClear(var_flags_to_clear[i]);}
	var_flags_to_clear.clear();
	assert(var_flags.allCleared());
}




TrBool BasicSolver::checkSubsumed(vector<Clause*>& clses,uint start){
    assert(decisionLevel() == 0);
    assert(start <= clses.size());
    vector<Clause*>::iterator beg = clses.begin() + start;
	vector<Clause*>::iterator to = beg;
	vector<Clause*>::iterator end = clses.end();
    while (beg!= end){
    	Clause* cls = *beg++;

    	if (clsCanBeOmitted(*cls)){
    		 statistics.nbr_simplify_worked += 1;
    		 clsUncreate(cls);
    		 continue;
    	}

    	assert(queue_begin == assign_stack.size());

        //if (!cls->isCardConstraint() && clsSubsumedByShort(*cls)){
    	if(clsSubsumedByShort(*cls)){
        	//print_out("Check subsumed worked for size %d \n",cls->size());
        	statistics.nbr_subsumed_worked += 1;
            clsUncreate(cls);
            if(!state_ok) return FALSE_TrBool;
        }else{
            *to++ = cls;
        }
    }
    if(to != end) clses.erase(to,end);
    return UNKNOWN_TrBool;
}



uint BasicSolver::clsFindSubsumed(Clause& cl){
	assert(!cl.isCardConstraint());

	uint new_sz = cl.size();
	Literal lit0, lit1,lit2;

	for(uint c = 0; c < cl.size();++c){
		if((lit0=cl[c],!var_flags.isSet(lit0.toVar()))) continue; //already deleted

		vector<WatchedElement>& watched	= litWatchedClauses(lit0);;

		for(vector<WatchedElement>::iterator iter =watched.begin();iter != watched.end();++iter){

			if(iter->isCardinality()){
				Clause& card_cls = *iter->getClause();
				for(uint k=card_cls.size();k--;){
					if((lit1=card_cls[k].flipped(),lit1==lit0 || !var_flags.isSet(lit1.toVar()))) continue;

					if(cl.contains(lit1)){
						assert(clsBinRedundant(lit0,lit1) && cl.contains(lit0) && cl.contains(lit1));
						return 0;} //subsumed by implicit binary lit0 lit1

					assert(clsBinRedundant(lit0,lit1) && cl.contains(lit0) && cl.contains(lit1.flipped()));
					var_flags.clear(lit1.toVar()); //shrink
					--new_sz;
				}
				continue;
			}


			if(!iter->isShort() || (lit1=iter->getFirstLiteral(),!var_flags.isSet(lit1.toVar()))) continue;

			if(iter->isBinary()){

				if(cl.contains(lit1)){
					assert(clsBinRedundant(lit0,lit1) && cl.contains(lit0) && cl.contains(lit1));
					return 0;} // subsumed by binary lit0 lit1

				assert(clsBinRedundant(lit0,lit1) && cl.contains(lit0) && cl.contains(lit1.flipped()));
				var_flags.clear(lit1.toVar());
				--new_sz;

			}else{
				if((lit2=iter->getSecondLiteral(),!var_flags.isSet(lit2.toVar()))) continue;
				bool l1_pol = cl.contains(lit1); assert(l1_pol || cl.contains(lit1.flipped()));
				bool l2_pol = cl.contains(lit2); assert(l2_pol || cl.contains(lit2.flipped()));

				if(!l1_pol && !l2_pol) continue;

				if(l1_pol && l2_pol){
					assert(clsTernRedundant(lit0,lit1,lit2) && cl.contains(lit0) && cl.contains(lit1)&& cl.contains(lit2) );
					return 0;} //subsumed by ternary lit0 lit1 lit2

				Literal del = l1_pol ? lit2 : lit1;
				assert(clsTernRedundant(lit0,lit1,lit2) && cl.contains(lit0) && cl.contains(del.flipped())&& cl.contains(l1_pol?lit1:lit2));
				var_flags.clear(del.toVar());
				--new_sz;
			}
		}
	}

	return new_sz;
}



uint	BasicSolver::clsFindSubsumedCardConstraint(Clause& cl){
	assert(cl.isCardConstraint());

	uint new_sz = cl.size();
	Literal lit0, lit1,lit2;

	for(uint c = 0; c < cl.size();++c){
		if((lit0=cl[c],!var_flags.isSet(lit0.toVar()))) continue; //already deleted

		vector<WatchedElement>& watched	= litWatchedClauses(lit0);;

		for(vector<WatchedElement>::iterator iter =watched.begin();iter != watched.end();++iter){

			if(!iter->isShort() || (lit1=iter->getFirstLiteral(),!var_flags.isSet(lit1.toVar()))) continue;

			if(iter->isBinary()){

				if(cl.contains(lit1)){// subsumed by binary lit0 lit1
					clsCreateBinary(lit0.flipped(),lit1.flipped());
					for(uint k=cl.size();k--;){
						assert(!litIsTrue(cl[k]));
						if(litIsUnknown(cl[k]) && cl[k] != lit1 && cl[k] != lit0){
							clsLearnUnit(cl[k].flipped());}
					}
					state_ok = (bcp(false) == BCP_SUCCESS);
					return 0;
				}

				assert(cl.contains(lit1.flipped()));
				var_flags.clear(lit1.toVar());
				--new_sz;
				assert(!litIsFalse(lit1));
				if(litIsUnknown(lit1)){	clsLearnUnit(lit1);}

			}else{
				if((lit2=iter->getSecondLiteral(),!var_flags.isSet(lit2.toVar()))) continue;
				bool l1_pol = cl.contains(lit1); assert(l1_pol || cl.contains(lit1.flipped()));
				bool l2_pol = cl.contains(lit2); assert(l2_pol || cl.contains(lit2.flipped()));

				if(!l1_pol && !l2_pol) continue;

				if(l1_pol && l2_pol){ //subsumed by ternary lit0 lit1 lit2
					clsCreateBinary(lit0.flipped(),lit1.flipped());
					clsCreateBinary(lit0.flipped(),lit2.flipped());
					clsCreateBinary(lit1.flipped(),lit2.flipped());

					for(uint k=cl.size();k--;){
						assert(!litIsTrue(cl[k]));
						if(litIsUnknown(cl[k]) && cl[k] != lit2 && cl[k] != lit1 && cl[k] != lit0){
							clsLearnUnit(cl[k].flipped());}
					}
					state_ok = (bcp(false) == BCP_SUCCESS);
					return 0;
				}

				Literal del = l1_pol ? lit2 : lit1;
				assert(cl.contains(del.flipped()));
				var_flags.clear(del.toVar());
				--new_sz;
				assert(!litIsFalse(del));
				if(litIsUnknown(del)){ clsLearnUnit(del);}
			}
		}
	}

	assert(new_sz > 0);
	return new_sz;
}

TrBool	BasicSolver::clsComputeClosure(const Literal& lit0,const we_vec&  shorts,const bool tern ){

	assert(lit0 != INVALID_LIT);
	if(varValue(lit0.toVar()).isKnown()) return UNKNOWN_TrBool;
	const Literal lit0_flp = lit0.flipped();
	Literal lit1,lit2;

	for(we_vec::const_iterator iter =shorts.begin();iter != shorts.end();++iter){
		assert(iter->isBinary() || iter->isTernary());

		lit1=iter->getFirstLiteral();
		if(lit1==lit0_flp || litIsTrue(lit1)) continue;
		assert(lit1 != INVALID_LIT && lit1.toVar().idx() < numberVariables());

		if(iter->isBinary()){
			if(lit0 == lit1 || litIsFalse(lit1)){
				assert(litIsUnknown(lit0));
				clsLearnUnit(lit0);
				state_ok = bcp(false) == BCP_SUCCESS;
				return state_ok ? TRUE_TrBool : FALSE_TrBool;

			}else if(!clsBinRedundant(lit0,lit1)){
				assert(litIsUnknown(lit0) && litIsUnknown(lit1));
				clsCreateBinary(lit0,lit1);
				++statistics.nbr_closure_create_bin;
			}
		}else if((lit2 = iter->getSecondLiteral(),lit2!=lit0_flp && !litIsTrue(lit2))){

			if(litIsFalse(lit1) && litIsFalse(lit2)){
				assert(litIsUnknown(lit0));
				clsLearnUnit(lit0);
				state_ok = bcp(false) == BCP_SUCCESS;
				return state_ok ? TRUE_TrBool : FALSE_TrBool;

			}else if(lit1 == lit0 || litIsFalse(lit1)){
				if(!clsBinRedundant(lit0,lit2)){
					assert(litIsUnknown(lit0) && litIsUnknown(lit2));
					clsCreateBinary(lit0,lit2);
					++statistics.nbr_closure_create_bin;}
			}else if(lit2==lit0 || litIsFalse(lit2)){
				if(!clsBinRedundant(lit0,lit1)){
					assert(litIsUnknown(lit0) && litIsUnknown(lit1));
					clsCreateBinary(lit0,lit1);
					++statistics.nbr_closure_create_bin;}
			}else if((tern || lit0.isPos() || lit1.isPos() || lit2.isPos()) && !clsTernRedundant(lit0,lit1,lit2)){
			//}else if(tern && !clsTernRedundant(lit0,lit1,lit2)){
				assert(litIsUnknown(lit0) && litIsUnknown(lit1) && litIsUnknown(lit2));
				clsCreateTernary(lit0,lit1,lit2);
				++statistics.nbr_closure_create_tern;
			}
		}
	}

	return UNKNOWN_TrBool;
}

void filterShorts(const we_vec& from, we_vec& to, bool& already_filtered){
	if(already_filtered) return;
	assert(to.empty());
	for(we_vec::const_iterator iter =from.begin();iter != from.end();++iter){
			if(iter->isShort()) to.push_back(*iter);
	}
	already_filtered = true;
}

TrBool	BasicSolver::clsComputeClosureLit(const Literal& lit){
	assert(lit != INVALID_LIT && litIsUnknown(lit));
	TrBool status 	= UNKNOWN_TrBool;
	we_vec& check 	= litWatchedClauses(lit);
	we_vec& cands	= litWatchedClauses(lit.flipped());
	we_vec  shorts;
	bool 	shorts_filtered = false;

	for(uint chk = 0; chk != check.size();++chk){

		WatchedElement& we = check[chk];
		if(we.isBinary()){
			filterShorts(cands,shorts,shorts_filtered);
			status = clsComputeClosure(we.getFirstLiteral(),shorts,statistics.nbr_closure_create_tern<settings.max_tern_closure );
			if(status.isKnown()) return status.isFalse()?FALSE_TrBool : UNKNOWN_TrBool;
		}else if(we.isCardinality()){
			assert(we.getClause() != NULL);
			Clause& card_cls = *(we.getClause());
			//this->printClause(*console_out,card_cls);
			assert(we.getClause()->isCardConstraint());

			for(uint k= card_cls.size();k--;){
				Literal impl = card_cls[k].flipped();
				if(impl == lit) continue;
				assert(clsBinRedundant(impl,lit));
				filterShorts(cands,shorts,shorts_filtered);
				status = clsComputeClosure(impl,shorts,statistics.nbr_closure_create_tern<settings.max_tern_closure );
				if(status.isKnown()) return status.isFalse()?FALSE_TrBool : UNKNOWN_TrBool;
			}
		}
	}
	return status;
}

TrBool	BasicSolver::clsComputeBinTernClosure(){

	assert(decisionLevel() == 0);
	assert(state_ok);
	assert(debugCheckCardWatchers());

	simplifyAllShortWatchers();

	TrBool status = UNKNOWN_TrBool;

	vector<Variable> vars_to_do;
	for(uint i = numberVariables();i--;){
		Variable v(i);
		if(act_hp.contains(v)) vars_to_do.push_back(v);}

	std::random_shuffle(vars_to_do.begin(), vars_to_do.end(),settings.randy);

	assert(debugCheckCardWatchers());

	for(uint i = vars_to_do.size();i--;){
		Variable v = vars_to_do[i];
		if(varValue(v).isKnown()) continue;

		assert(debugCheckCardWatchers(litWatchedClauses(v.toPosLit())));
		status = clsComputeClosureLit(v.toPosLit());
		if(status.isKnown()) break;

		if(varValue(v).isKnown()) continue;

		assert(debugCheckCardWatchers(litWatchedClauses(v.toNegLit())));
		status = clsComputeClosureLit(v.toNegLit());
		if(status.isKnown()) break;
	}

	return status;
}



bool BasicSolver::clsSubsumedByShort(Clause& cl){
	assert(var_flags.allCleared());
	assert(decisionLevel() == 0);
	assert(queue_begin == assign_stack.size());

	for(uint c = cl.size();c--;){var_flags.setFlag(cl[c].toVar());}

	uint new_sz = cl.isCardConstraint() ? clsFindSubsumedCardConstraint(cl): clsFindSubsumed(cl);

	//assert(!cl.isCardConstraint());
	//uint new_sz = clsFindSubsumed(cl);

	if(new_sz == cl.size() || new_sz == 0){
		for(uint c = cl.size();c--;){var_flags.carelessClear(cl[c].toVar());}
		return new_sz == 0;
	}

	if(cl.isCardConstraint()) {statistics.nbr_subsumed_shrink_card += (cl.size() - new_sz);}else {
	statistics.nbr_subsumed_shrinked += (cl.size() - new_sz);}



	//print_out("Check subsumed shrinks to %d from %d \n",new_sz,cl.size());
	if(new_sz < 4){

		uint to = 0;
		Literal new_lits[new_sz];
		for(uint i=0;i!=cl.size();++i){
			Literal l = cl[i];
			if(var_flags.isSet(l.toVar())){ assert(to<new_sz);
				new_lits[to++] = l;	var_flags.clear(l.toVar());	}
		}

		assert(new_sz > 0);
		if(new_sz == 3){
			if(!clsTernRedundant(new_lits[0],new_lits[1],new_lits[2])){
				clsCreateTernary(new_lits[0],new_lits[1],new_lits[2]);}
			if(cl.isCardConstraint()){
				clsCreateBinary(new_lits[0].flipped(),new_lits[1].flipped());
				clsCreateBinary(new_lits[0].flipped(),new_lits[2].flipped());
				clsCreateBinary(new_lits[1].flipped(),new_lits[2].flipped());
			}
		}else if(new_sz == 2){
			if(!clsBinRedundant(new_lits[0],new_lits[1])){
				clsCreateBinary(new_lits[0],new_lits[1]);}
			if(cl.isCardConstraint()){clsCreateBinary(new_lits[0].flipped(),new_lits[1].flipped());}
		}else {clsLearnUnit(new_lits[0]);}

		state_ok = (bcp(false) == BCP_SUCCESS);
		return true;
	}


	bool unlink = !var_flags.isSet(cl[0].toVar()) || !var_flags.isSet(cl[1].toVar()) || cl.isCardConstraint();

	if(unlink){clsUnlinkWatched(&cl);}

	uint to = 0;
	for(uint i=0;i!=cl.size();++i){
		Literal l = cl[i];
		if(var_flags.isSet(l.toVar())){	cl[to++] = l;	var_flags.clear(l.toVar());	}
	}
	assert(to == new_sz);
	clsShrinkToSize(cl,to);

	if(unlink){clsLinkWatched(&cl);}
	state_ok = (bcp(false) == BCP_SUCCESS);
	return false;
}







TrBool 	BasicSolver::removeSatAndShrinkClauses(vector<Clause*>& clses,uint32_t max_cls_sz){

	TrBool status = UNKNOWN_TrBool;

	vector<Clause*>::iterator to, iter;
	for (iter = to = clses.begin(); iter != clses.end(); ++iter){


		Clause& cl = **iter;
		/*if(cl.isCardConstraint()){
			*to++ = *iter;
			continue;
		}*/

		if(cl.size() > max_cls_sz){
			if (clsCanBeOmitted(cl)){
				statistics.nbr_simplify_worked += 1;
				clsUncreate(*iter);
			}else{
				*to++ = *iter;
			}

			continue;
		}

		uint32_t new_cls_sz = clsCheckSatAndShrink(cl);
		//print_out("shrink from %d to %d \n",cl.size(),new_cls_sz);
		state_ok = bcp(false) == BCP_SUCCESS;
		if(!state_ok ){status = FALSE_TrBool; break;}

		if(new_cls_sz <= 3){

			if(new_cls_sz == 3){
				assert(litIsUnknown(cl[0]) && litIsUnknown(cl[1]) && litIsUnknown(cl[2]));
				clsCreateTernary(cl[0],cl[1],cl[2]);
				if(cl.isCardConstraint()){
					clsCreateBinary(cl[0].flipped(),cl[1].flipped());
					clsCreateBinary(cl[0].flipped(),cl[2].flipped());
					clsCreateBinary(cl[1].flipped(),cl[2].flipped());
				}
			}else if(new_cls_sz == 2){
				assert(litIsUnknown(cl[0]) && litIsUnknown(cl[1]));
				clsCreateBinary(cl[0],cl[1]);
				if(cl.isCardConstraint()){
					clsCreateBinary(cl[0].flipped(),cl[1].flipped());
				}
			}else if(new_cls_sz == 1){

				if(litIsUnknown(cl[0])){
					clsLearnUnit(cl[0]);
					state_ok = bcp(false) == BCP_SUCCESS;
				}

				if(!state_ok || litIsFalse(cl[0])){
					while(iter != clses.end()){*to++ =*iter++;}
					status = TrBool(state_ok = false);
					break;
				}
			}

			statistics.nbr_simplify_worked += 1;
			clsUncreate(&cl,false);
		}else{
			if(new_cls_sz < cl.size()){
				clsShrinkToSize(cl,new_cls_sz);
				clsLinkWatched(&cl);
			}
			*to++ = &cl;
		}
	}


	clses.erase(to,clses.end());

	return status;
}





//return 0 means that clause is not needed any longer (may be unlinked)
//return cl.size() means that clause has to remain unchanged
//return else -> clause can be shrinked to this size (clause is unlinked in this case)
uint32_t BasicSolver::clsCheckSatAndShrink(Clause& cl){

	assert(decisionLevel() == 0);

	if(litIsTrue(cl[0])){
		clsUnlinkWatched(&cl);
		return 0;
	}

	const uint32_t q_begin = assign_stack.size();
	assert(q_begin == queue_begin);

	uint32_t new_end = cl.size();
	bool cls_linked = true;

	vector<Literal> removed_lits;

	for(uint32_t i = cl.size(); --i;) {
		Literal lit = cl[i];

		if(litIsFalse(lit)){
			if(cls_linked){clsUnlinkWatched(&cl); cls_linked = false;}
			if(cl.isCardConstraint()) removed_lits.push_back(lit);
			cl[i]		= cl[--new_end];
			assert(litIsFalse(cl[i]));
			continue;
		}

		if(litIsTrue(lit) || (uncheckedEnqueue<Clause*>(lit.flipped()),bcp(false) != BCP_SUCCESS)){

			if(cls_linked){clsUnlinkWatched(&cl);}

			backjumpUndoAssigns(q_begin,BACKJUMP_PREV_UNTOUCHED);

			if(cl.isCardConstraint()){
				for(uint k=removed_lits.size();k--;){
					assert(litIsUnknown(removed_lits[k]));
					uncheckedEnqueue<Clause*>(removed_lits[k].flipped());}
				for(uint k=i;k--;){
					assert(litIsUnknown(cl[k]));
					uncheckedEnqueue<Clause*>(cl[k].flipped());}

			}else if(litIsTrue(lit)){return 0;}

			uint32_t c_end = 0;
			while( i != new_end){cl[c_end++] = cl[i++];}

			statistics.nbr_bigsimplify_shrinked += (cl.size() - c_end);
			return c_end;
		}
	}

	backjumpUndoAssigns(q_begin,BACKJUMP_PREV_UNTOUCHED);

	if(new_end < cl.size()){
		if(cl.isCardConstraint()) {statistics.nbr_big_simp_card_shrink += (cl.size() - new_end);}
		else{ statistics.nbr_bigsimplify_shrinked += (cl.size() - new_end);}
	}

	return  new_end;
}


void 	BasicSolver::printClause(OFStream& out,  Clause& cls){

	//assert(decisionLevel() == 0);
	Literal lits[cls.size()];
	uint sz = 0;
	for(uint32_t l= cls.size(); l--;){
		Literal lit = cls[l];
		if(litIsTrue(lit)) return;
		if(litIsUnknown(lit)){lits[sz++] = lit;}
	}


	while(sz--){
		out.writeString("%d ",lits[sz].toDIMACS() );
	}
	if(cls.isCardConstraint())	out.writeString(" 0# \n");
	else						out.writeString(" 0 \n");
}

void	BasicSolver::printDIMACS(OFStream* out, OFStream* learnts_out){

	if(decisionLevel() != 0){
		out->writeString("c !!! THIS CNF FILE WAS CREATED AT DECISION LEVEL %d \n",decisionLevel());}

	//unit clauses
	for(uint i = 0; i != assign_stack.size(); ++i){
		out->writeString("%d 0\n",assign_stack[i].toDIMACS());
	}

	//long clauses
	for(uint32_t i = 0; i != fix_cls.size(); ++i){
		assert(fix_cls[i] != NULL);
		printClause(*out,*fix_cls[i]);
	}

	//short clauses
	for(uint i = this->numberVariables()*2; i--;){
		Literal lit(i);

		we_vec& watches = litWatchedClauses(lit);

		for(uint j = 0; j != watches.size(); ++j){
			WatchedElement& we = watches[j];
			if(!we.isShort()) continue;

			Literal lit2 = we.getFirstLiteral();
			if(lit.idx() > lit2.idx()) continue;

			if(we.isBinary()){
				out->writeString("%d %d 0 \n",lit.toDIMACS(),lit2.toDIMACS());
			}else{

				Literal lit3 = we.getSecondLiteral();
				if(lit.idx() > lit3.idx()) continue;
				out->writeString("%d %d %d 0 \n",lit.toDIMACS(),lit2.toDIMACS(),lit3.toDIMACS());
			}
		}
	}


	if(learnts_out == NULL){ return;}

	learnts_out->writeString("l \t \t \t \n");
	for(uint32_t i = 0; i != learnt_cls.size(); ++i){
		assert(learnt_cls[i] != NULL);
		printClause(*learnts_out,*learnt_cls[i]);
	}

}

//arc consistency
TrBool	BasicSolver::checkCardConstraints(){
	assert(state_ok);
	assert(queue_begin == assign_stack.size());

	TrBool	result					= UNKNOWN_TrBool;
	uint	start_level				= decisionLevel();
	uint	start_assigns			= assign_stack.size();

	while(true){

		int32_t bcp_res	=	bcpCardConstraints();

		if(bcp_res == BCP_SUCCESS){		//there is no conflict

			Literal next	= nextDecision(Polarity_Prev);
			if (next == INVALID_LIT){
				backjumpKeepLevel(start_level,BACKJUMP_SET_PREV);
				queue_begin = start_assigns; //to propagate changes at this decision level
				//backjumpUndoAssigns(start_stack_sz,BACKJUMP_SET_PREV);
				return TRUE_TrBool;
			}

			// Increase decision level and enqueue 'next'
			assert(litIsUnknown(next));
			newDecisionLevel();
			uncheckedEnqueue<Clause*>(next);


		}else{						//there is a conflict

			++statistics.nbr_conflicts;
			assert(bcp_res != 1);

			if (!firstUIP(bcp_res))  {
				print_out("Determine unsat in card constraint check \n");
				return TrBool(state_ok = false);
			}

			assert(litAssignLevel(conflict_lemma[0]) == decisionLevel());

			reduceSelfImpl();

			uint32_t backtrack_level = backtrackLevel();
			backjumpKeepLevel(backtrack_level,BACKJUMP_SET_PREV);

			assert(litIsUnknown(conflict_lemma[0]));

			const uint32_t l_sz(conflict_lemma.size());
			if(l_sz > 3){
				assert(l_sz <= Clause::MAX_SIZE);
				Clause* learnt_cl = clsCreate(conflict_lemma.begin(),conflict_lemma.end(),true,false);
				if(learnt_cl == NULL){break;}
				clsBump(learnt_cl);
				uncheckedEnqueue(conflict_lemma[0], learnt_cl);
			}else if(l_sz == 3){
				assert(!clsTernRedundant(conflict_lemma[0],conflict_lemma[1],conflict_lemma[2]));
				clsCreateTernary(conflict_lemma[0],conflict_lemma[1],conflict_lemma[2]);
				uncheckedEnqueue(conflict_lemma[0],conflict_lemma[1],conflict_lemma[2]);
			}else if(l_sz == 2){
				assert(!clsBinRedundant(conflict_lemma[0],conflict_lemma[1]));
				clsCreateBinary(conflict_lemma[0],conflict_lemma[1]);
				uncheckedEnqueue(conflict_lemma[0],conflict_lemma[1]);
			}else{
				clsLearnUnit(conflict_lemma[0]);
			}

			if(backtrack_level < start_level){break;}
		}
	}

	while(bcp(false) != BCP_SUCCESS){
		if(decisionLevel() == 0){return TrBool(state_ok = false);}
		backjumpKeepLevel(decisionLevel()-1,BACKJUMP_SET_PREV);
	}

	statistics.nbr_arc_const_levels_omit += (start_level - decisionLevel());
	//print_out("Arc consistency failed at level %d back to level %d \n",start_level,decisionLevel());
	return UNKNOWN_TrBool;

}


int32_t	BasicSolver::bcpCardConstraints(){
	int32_t			nbr_pivot_lits = BCP_SUCCESS; //indicates no conflict by now
	uint32_t		num_props(0);

	while(queue_begin < assign_stack.size()){

		++num_props;
		nbr_pivot_lits = bcpCardConstPropagateLit(assign_stack[queue_begin++].flipped());
	}

	statistics.nbr_propagations	+= num_props;
	return nbr_pivot_lits;
}

int32_t	BasicSolver::bcpCardConstPropagateLit(const Literal& false_lit){
	we_vec& watched	= litWatchedClauses(false_lit); //watched clauses that have to be checked when lit gets false
	we_iter iter = watched.begin();
	we_iter copy = iter;
	we_iter end  = watched.end();

	while(iter != end){
		WatchedElement& we = *iter++;

		if(we.isShort()){
			*copy++ = we; //keep this watcher but ignore clauses
			continue;
		}

		if(we.isCardinality()){
			*copy++ = we; //keep this watcher
			Clause& card_cls = *we.getClause();
			assert(card_cls.isCardConstraint());

			for(uint k= card_cls.size();k--;){
				Literal impl = card_cls[k].flipped();
				if(impl==false_lit) continue; //card_cls[k] is the one to be true

				//implied binary clause
				if(litIsUnknown(impl)){
					uncheckedEnqueue<const Literal&>(impl,false_lit);
				}else if(litIsFalse(impl)){
					bcpFinishOnConflict(watched,iter,copy,end);
					return bcpInitConflict(decisionLevel(),false_lit,impl);
				}
			}
			continue;
		}


		//might be any other literal of this clause except the watched literal
		const Literal other = we.getFirstLiteral();
		assert(other != false_lit);

		//keep this watcher if other lit is true
		if(litIsTrue(other)){ *copy++ = we; continue;}

		Clause& cl = *(we.getClause());
		if(!cl.isCardConstraint()){
			*copy++ = we; //keep this watcher but ignore clauses
			continue;
		}

		uint32_t other_watcher_index = (cl[1] != false_lit);
		assert(cl[other_watcher_index ^ 1] == false_lit && cl[other_watcher_index] != false_lit);

		//it might be that the shortcut literal in 'we' does not point to the actual watched literal anymore
		if(litIsTrue(cl[other_watcher_index])){
			assert(!we.isShort());
			we.setFirstLiteral(cl[other_watcher_index]);
			*copy++ = we; //keep this watcher
			continue;
		}

		if(bcpFindNewWatcher(cl,false_lit,other_watcher_index)){continue;}


		// No new watched found -- clause is unit under assignment:
		we.setFirstLiteral(cl[other_watcher_index]);
		*copy++				= we;

		if (litIsFalse(cl[other_watcher_index])){
			if(cl.isLearnt())clsBump(&cl);

			bcpFinishOnConflict(watched,iter,copy,end);

			const int32_t dec_level = decisionLevel();
			int32_t  nbr_pivot_lits = bcpInitConflict(dec_level,false_lit,cl[other_watcher_index]);
			for(uint32_t k = 2;k<cl.size();++k){firstUIPCheckLiteral(cl[k],dec_level,nbr_pivot_lits);}
			assert(nbr_pivot_lits != 1);
			return nbr_pivot_lits;

		}else{
			uncheckedEnqueue(cl[other_watcher_index], &cl);
		}

	}
	if(copy != end) watched.erase(copy,end);

	return BCP_SUCCESS;
}








const char*  basic_solver_cpp_version(){return "BasicSolver.cpp:    $Revision: 1.12 $";}
