/*
 Stephan Kottler
 kottlers@informatik.uni-tuebingen.de
 *
 */


#ifndef TYPES_HPP_
#define TYPES_HPP_

#include <vector>
#include <cassert>
#include <stdio.h>
#include "LazyFraction.hpp"
#include "OFStream.hpp"
#include "Settings.hpp"


using std::vector;

#ifndef NO_CLAUSE_ACT
#define CLAUSE_ACT 1
#endif


class Literal;
class Clause;

const uint32_t zero32 = 0;
const uint32_t zero64 = 0;

/**Class to represent boolean values and a third state 'Unknown'*/
class TrBool{
	uint8_t val;

//public members
public:
	inline			TrBool();

	/**Constructs an object with the specified value.
	* @param i: a value from {0,1,2}*/
	inline explicit	TrBool(uint8_t i);

	inline			TrBool(const TrBool& b);

	inline uint8_t	toInt()							const;
	inline bool		isTrue()						const;
	inline bool		isFalse()						const;
	inline bool		isUnknown()						const;
	inline bool		isKnown()						const;
	inline bool		operator==(const TrBool& o)		const;
	inline bool		operator!=(const TrBool& o)		const;
};


/**Class to represent variables. All additional information about a variable are stored outside: in an instance of
 * ConstraintDB*/
class Variable{

	uint32_t ID;

public:
	inline explicit	Variable(const uint32_t i);

	inline uint32_t	idx()							const;
	inline Literal	toPosLit()						const;
	inline Literal	toNegLit()						const;
	inline uint		toDIMACS()						const;
	inline bool		operator<(const Variable& v)	const;
	inline bool		operator!=(const Variable& v)   const;
	inline bool		operator==(const Variable& v)   const;

	friend class Literal;
	friend class ConstraintDB;
};


/**Class to represent literals. All aditional information are stored elsewhere: in an instance of class ConstraintDB*/
class Literal{

	uint32_t ID;

public:
	/**Return a literal object from the given DIMACS number*/
	inline static Literal DIMACS2Literal(int32_t d);

	/**Construct an invalid literal*/
	inline				Literal();

	/**Construct a literal with the specified id*/
	inline explicit		Literal(const uint32_t i);

	inline uint32_t		idx()			const;
	inline Literal		flipped()		const;
	inline Variable		toVar()			const;
	inline bool			isPos()			const;
	inline bool			isNeg()			const;
	inline TrBool		pol()			const;
	inline TrBool		flpdPolarity()	const;
	inline int			toDIMACS()		const;

	/**Compare literals by their ids*/
	inline bool			operator<(const Literal& l) const;
	inline bool			operator!=(const Literal& l) const;
	inline bool			operator==(const Literal& l)const;

	friend class Variable;
	friend class ConstraintDB;
};


/**Class to represent clauses. Contrary to the other types each clause is represented by exactly one instance of this class Clause.
 * Hence, we deal with pointers of this class. Note that all instances of this class can have different sizes although sizeof(Clause)
 * will always return the same value.*/
class Clause{

public:
	static const uint32_t MAX_SIZE 	= ((~zero32)>> 1) - 2;
	static const uint32_t MAX_STAMP = (1 << 20) - 1;

private:

#ifdef CLAUSE_ACT
	cls_act_t activity;
#endif
	uint32_t size30_card_learnt;

	//uint32_t nextk12_stamp20;

	/**The literals start here. The fields are filled from outside the class.*/
	Literal literals[0];

	/**Construct a part of the clause.*/
	inline Clause(const uint32_t& sz, bool learnt, bool cardinality);

	/**Copying of Clause instances is not allowed*/
#ifdef CLAUSE_ACT
	Clause(const Clause& cl):activity(cl.activity){assert(false);}
#else
	Clause(const Clause& cl){assert(false);}
#endif
	Clause& operator=(const Clause& c){assert(false);return (Clause&)c;}

	inline void				removeAt(uint32_t idx);
	inline void				shrinkToSize(uint32_t new_sz);

public:
	struct RelevanceDescendingFunctor{
		bool operator() (Clause* a, Clause* b){	costly_assert(a->isLearnt() && b->isLearnt());
#ifdef CLAUSE_ACT
		return a->activity  > b->activity;
#else
		return a->size()  < b->size();
#endif
	}};

	/**Access the literals of a clause by an index*/
	inline Literal&			operator [] (uint32_t i);

	inline const uint32_t	size()				const;
	inline const bool		contains(const Literal& l)	const;
	inline const bool		isLearnt()			const;
	inline const bool		isCardConstraint()	const;

	/*inline uint32_t 		getNextK(const uint32_t& stamp);
	inline void 			setNextK(const uint32_t k, const uint32_t& stamp);
	inline void				clearStamp();*/


	friend class ConstraintDB;
};



//Implementation
//>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
const Literal   INVALID_LIT			(~zero32);
const Variable  INVALID_VAR			(~zero32);

const uint8_t   FALSE_TRBOOL_VAL	= 0;
const uint8_t   TRUE_TRBOOL_VAL	= 1;
const uint8_t   TRBOOL_FLIP_MASK	= 1;
const uint8_t   UNKNOWN_TRBOOL_VAL	= 2;
const TrBool	TRUE_TrBool(TRUE_TRBOOL_VAL);
const TrBool	FALSE_TrBool(FALSE_TRBOOL_VAL);
const TrBool	UNKNOWN_TrBool(UNKNOWN_TRBOOL_VAL);



//TBool Implementation
			TrBool::TrBool()						:val(UNKNOWN_TRBOOL_VAL){ }
			TrBool::TrBool(const TrBool& b)			:val(b.val)			{ }
			TrBool::TrBool(uint8_t i)				:val(i)	{
							assert(i == TRUE_TRBOOL_VAL || i == FALSE_TRBOOL_VAL || i == UNKNOWN_TRBOOL_VAL);}

uint8_t		TrBool::toInt()				const	{return val;}
bool		TrBool::isTrue()			const	{return val == TRUE_TRBOOL_VAL;}
bool		TrBool::isFalse()			const	{return val == FALSE_TRBOOL_VAL;}
bool		TrBool::isUnknown()			const	{return val == UNKNOWN_TRBOOL_VAL;}
bool		TrBool::isKnown()			const	{return val != UNKNOWN_TRBOOL_VAL;}
bool		TrBool::operator==(const TrBool& o) const   {return val == o.val;}
bool		TrBool::operator!=(const TrBool& o) const   {return val != o.val;}



//Variable Implementation
			Variable::Variable(const uint32_t i) :ID(i)		{ }
Literal		Variable::toPosLit()					const   {return Literal((ID << 1) + 1);}
Literal		Variable::toNegLit()					const   {return Literal(ID << 1);}
uint32_t	Variable::idx()							const   {return ID;}
uint32_t	Variable::toDIMACS()					const   {return ID+1;}
bool		Variable::operator<(const Variable& v)  const   {return ID < v.ID;}
bool		Variable::operator!=(const Variable& v) const   {return ID != v.ID;}
bool		Variable::operator==(const Variable& v) const   {return ID == v.ID;}

//Literal Implementation
			Literal::Literal()					:ID(~zero32){ }
			Literal::Literal(const uint32_t i)   	:ID(i)	{ }
uint32_t	Literal::idx()							const   {return ID;}
Literal		Literal::flipped()						const   {return Literal(ID ^ 1);}
Variable	Literal::toVar()						const   {return Variable(ID >> 1);}
bool		Literal::isPos()						const   {return ID & 1;}
bool		Literal::isNeg()						const   {return !(ID & 1);}
int			Literal::toDIMACS()						const   {return ID&1 ? (ID>>1)+1 : -(ID>>1)-1;}
TrBool		Literal::pol()							const   {return TrBool(ID & 1);}
TrBool		Literal::flpdPolarity()					const   {return TrBool(!(ID&1));}
bool		Literal::operator==(const Literal& l)   const   {return ID == l.ID; }
bool		Literal::operator<(const Literal& l)	const   {return ID < l.ID;}
bool		Literal::operator!=(const Literal& l)   const   {return ID != l.ID;}

Literal	Literal::DIMACS2Literal(int32_t d){
	assert(d != 0);
	//return Literal(d < 0 ? (-d-1)*2 : d*2-1);
	uint32_t m(d>>31);
	return Literal(((d^m)<<1)-(m+1));
}


const uint32_t  Clause::size()				const   {return size30_card_learnt >> 2;}
const bool		Clause::isLearnt()			const   {return size30_card_learnt & 1;}
const bool		Clause::isCardConstraint()	const	{return size30_card_learnt & 2;}
Literal&		Clause::operator [] (uint32_t i)	{assert(i<size());return literals[i];}

const bool		Clause::contains(const Literal& l)	const{
	for(uint i=size();i--;){if(literals[i]==l) return true;}
	return false;
}

void			Clause::removeAt(uint32_t idx){
	assert(idx > 1 && idx < size());
	literals[idx] = literals[size() - 1];
	size30_card_learnt -= 4;
	//clearStamp();
}

void			Clause::shrinkToSize(uint32_t new_sz){
	assert(new_sz > 3 && new_sz < size());

	size30_card_learnt = (new_sz<< 2) | (size30_card_learnt & 3) ;
	//clearStamp();
	assert(size() == new_sz);
}

Clause::Clause(const uint32_t& sz, bool learnt,bool cardinality):
#ifdef CLAUSE_ACT
	activity(0),
#endif
	size30_card_learnt( (sz << 2) |(((uint32_t)cardinality)<<1) |((uint32_t)learnt)){
	assert(cardinality == isCardConstraint());
	assert(learnt == isLearnt());
}


/*
Clause::Clause(const uint32_t& sz, bool learnt):
	activity(0),size31_learnt( (sz << 1) | ((uint32_t)learnt)),nextk12_stamp20(0){}


uint32_t Clause::getNextK(const uint32_t& stamp){
	assert(stamp <= MAX_STAMP);
	const bool valid = (nextk12_stamp20 & MAX_STAMP) == stamp;
	return (valid * (nextk12_stamp20 >> 20))+ 2;
}

void 	Clause::setNextK(const uint32_t k, const uint32_t& stamp){
	assert(stamp <= MAX_STAMP);
	assert(k >= 2);
	nextk12_stamp20 = ((k-2) << 20) | stamp;
}

void	Clause::clearStamp(){nextk12_stamp20 = 0;}
*/





#define types_hpp_version "Types.hpp:       $Revision: 1.1 $"

#endif
